/* SciPy specifics */

/* Version switcher colors from PyData Sphinx Theme */

.version-switcher__button[data-active-version-name*="dev"] {
  background-color: var(--pst-color-warning);
  border-color: var(--pst-color-warning);
  opacity: 0.9;
}

.version-switcher__button:not([data-active-version-name*="stable"]):not([data-active-version-name*="dev"]):not([data-active-version-name*="pull"]) {
  background-color: var(--pst-color-danger);
  border-color: var(--pst-color-danger);
  opacity: 0.9;
}

button.btn.version-switcher__button, button.btn.version-switcher__button:hover {
  color: black;
}

/* Main index page overview cards */

.sd-card .sd-card-img-top {
  height: 60px;
  width: 60px;
  margin-left: auto;
  margin-right: auto;
  margin-top: 10px;
}

/* Main index page overview images */

html[data-theme=dark] .sd-card img[src*='.svg'] {
  filter: invert(0.82) brightness(0.8) contrast(1.2);
}

/* Legacy admonition */

div.admonition-legacy {
  border-color: var(--pst-color-warning);
}

.admonition>.admonition-title::after,
div.admonition>.admonition-title::after {
  color: var(--pst-color-warning);
}

.admonition>.admonition-title,
div.admonition>.admonition-title {
  background-color: var(--pst-color-warning-bg);
}

/* JupyterLite "Try Examples" directive */

.try_examples_button {
    color: var(--pst-color-background);
    background-color: var(--pst-color-primary);
    border: none;
    padding: 5px 10px;
    border-radius: 10px;
    margin-bottom: 5px;
    box-shadow: 0 2px 5px var(--pst-color-surface);
    font-weight: bold;
    font-size: small;
}

.try_examples_button:hover {
  background-color: var(--pst-color-primary);
  transform: scale(1.02);
  opacity: 0.9;
  box-shadow: 0 2px 5px var(--pst-color-surface);
  cursor: pointer;
}

.try_examples_button_container {
    display: flex;
    justify-content: flex-end;
    gap: 10px;
    margin-bottom: 20px;
}

/* Wrap long titles in pages */
h1 {
  word-wrap: break-word;
}

/* Monospace titles for API docs */
div.empty + section>h1 {
  font-family: var(--pst-font-family-monospace);
}

.prename {
  font-family: var(--pst-font-family-monospace);
  font-size: var(--pst-font-size-h4);
}

.sig-prename {
  display: none;
}
